/*******************************************************************************/
/* debug.h - peephole debug header file 				       */
/*  v. 0.1 								       */
/*  Copyright (C) 2004 Alessandro Di Marco - All rights reserved 	       */
/*  by Alessandro Di Marco <dmr@gmx.it>	                 		       */
/* 									       */
/* This program is free software; you can redistribute it and/or 	       */
/* modify  under the terms of the GNU General Public License as 	       */
/* published by the Free Software Foundation; either version 2, or (at 	       */
/* your option) any later version.					       */
/* 									       */
/* This program is distributed in the hope that it will be useful, but 	       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of 		       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 	       */
/* General Public License for more details.				       */
/* 									       */
/* You should have received a copy of the GNU General Public License 	       */
/* along with GNU Emacs; see the file COPYING.  If not, write to the 	       */
/* Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.	       */
/*******************************************************************************/

/*
 *  $Header: /var/cvsroot/peephole/src/debug.h,v 1.3 2004/04/18 20:58:53 dmr Exp $
 */

#ifndef DEBUG_H
#define DEBUG_H

#ifdef DEBUG

#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include <stdio.h>

#define DEBUG_FILE "/var/log/peephole.log"

#define FIX_ERRNO(code) do { int err = errno; code; errno = err; } while(0)

#define PRINT(fmt, args...) FIX_ERRNO( do {		       	\
		FILE *_debfd;					\
								\
		_debfd = fopen(DEBUG_FILE, "a");		\
		if (!_debfd) {					\
			usleep(5000);				\
			continue;				\
		}						\
								\
		fprintf(_debfd , fmt , ## args);		\
		fclose(_debfd);					\
	} while(0) )

#define PRINTD(fmt, args...) FIX_ERRNO( do {			\
		time_t tt = time(NULL);				\
		struct tm tm;					\
		char buf[80];					\
								\
		buf[strlen(ctime_r(&tt, buf)) - 1] = '\0';	\
								\
		PRINT("[%s / %p] " fmt ,			\
		      buf , pthread_self(), ## args);		\
	} while(0) )
#else

#define PRINT(fmt, args...) do {} while(0)
#define PRINTD(fmt, args...) do {} while(0)

#endif /* DEBUG */

#endif /* DEBUG_H */
