/*
 *  $Header: /var/cvsroot/peephole/src/grammar/grammar.c,v 1.4 2004/04/18 20:58:53 dmr Exp $
 *
 *  Copyright (C) 2004 Alessandro Di Marco
 */

/*
 *  TODO:
 */

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "grammar.h"

#define FALSE 0
#define TRUE (!FALSE)

static const char cvsid[]="$Id: grammar.c,v 1.4 2004/04/18 20:58:53 dmr Exp $";

void signaler(char *string, void *arg)
{
	struct wm_providers *wmp = arg;
	struct wm_local *wml = wmp->root;
	int level = wmp->level;

	struct list_head *pos1, *pos2, *n;

	list_for_each_safe(pos1, n, &wml->watches.off[level]) {
		struct wm_watches *wmw = list_entry(pos1, struct wm_links, others[level])->watch;
		int turnon = TRUE;

		list_for_each(pos2, wmw->keywords) {
			struct wm_keywords *wmk = list_entry(pos2, struct wm_keywords, others);

			if (regexec(&wmk->compiled, string, 0, NULL, 0)) {
				turnon = FALSE;
				break;
			}
		}

		if (turnon) {
			list_del(&wmw->overlay->others[level]);

			(void) pthread_mutex_lock(&wml->freeze[level]);

			list_add(&wmw->overlay->others[level], &wml->watches.on[level]);

			(void) pthread_mutex_unlock(&wml->freeze[level]);
		}
	}
}

void init_overlay(struct wm_local *wml, int level)
{
	struct list_head *pos;

	(void) pthread_mutex_lock(&wml->freeze[level]);

	INIT_LIST_HEAD(&wml->watches.on[level]);

	(void) pthread_mutex_unlock(&wml->freeze[level]); 
  
	INIT_LIST_HEAD(&wml->watches.off[level]);

	list_for_each(pos, &wml->watches.main) {
		struct wm_watches *wmw = list_entry(pos, struct wm_watches, others);

		list_add(&wmw->overlay->others[level], &wml->watches.off[level]);
	}
}

struct wm_local *parser(char *path, char *name)
{
	void *config;
	struct wm_grammar_stub *hooks;

	struct wm_local *wml;

	config = dlopen(LIBPATH  "/local.so", RTLD_NOW);
	if (!config) {
		printf("%s\n", dlerror());
		return NULL;
	}

	hooks = (struct wm_grammar_stub *) dlsym(config, "hooks");
	if (!hooks) {
		printf("%s\n", dlerror());
		dlclose(config);
		return NULL;
	}

	wml = hooks->parse(path, name);

	dlclose(config);

	if (wml) {
		struct list_head *pos1, *pos2;
		int i;

		wml->freeze = (pthread_mutex_t *) malloc(wml->grade * sizeof (pthread_mutex_t));
		if (!wml->freeze) {
			free(wml);
			return NULL;
		}
		
		wml->watches.on = (struct list_head *) malloc(wml->grade * sizeof (struct list_head));
		if (!wml->watches.on) {
			free(wml->freeze);
			free(wml);
			return NULL;
		}

		wml->watches.off = (struct list_head *) malloc(wml->grade * sizeof (struct list_head));
		if (!wml->watches.off) {
			free(wml->watches.on);
			free(wml->freeze);
			free(wml);
			return NULL;
		}

		for (i = 0; i < wml->grade; i++) {
			(void) pthread_mutex_init(&wml->freeze[i], NULL);

			INIT_LIST_HEAD(&wml->watches.on[i]);
			INIT_LIST_HEAD(&wml->watches.off[i]);
		}

		(void) pthread_mutex_init(&wml->sentry, NULL);

		INIT_LIST_HEAD(&wml->coils);
		INIT_LIST_HEAD(&wml->clients);

		list_for_each(pos1, &wml->watches.main) {
			struct wm_watches *wmw = list_entry(pos1, struct wm_watches, others);

			wmw->overlay = (struct wm_links *) malloc(sizeof (struct wm_links) + wml->grade * sizeof (struct list_head));
			if (!wmw->overlay) {
				/* FIXME: some deallocations are needed here! */
				return NULL;
			}

			wmw->overlay->watch = wmw;

			list_for_each(pos2, wmw->keywords) {
				struct wm_keywords *wmk = list_entry(pos2, struct wm_keywords, others);

				if (regcomp(&wmk->compiled, wmk->pattern, REG_EXTENDED|REG_NOSUB|REG_NEWLINE)) {
					printf("cannot compile regexp!\n");
				}
			}
		}
	}

	return wml;
}
