/*
 *  $Header: /var/cvsroot/peephole/src/grammar/grammar.h,v 1.8 2004/04/19 10:20:30 dmr Exp $
 *
 *  Copyright (C) 2004 Alessandro Di Marco
 */

#ifndef GRAMMAR_H
#define GRAMMAR_H

#include <regex.h>
#include <netdb.h>
#include <pthread.h>
#include <sys/types.h>

#include "../peephole.h"
#include "../providers/providers.h"
#include "../list.h"

struct wm_local {
	int shutdown;
	
	struct list_head others;
	char *user;
	char *pass;

	pthread_mutex_t *freeze;

	struct list_head providers;
	int grade; // the amount of providers in providers list
	struct {
		struct list_head main;
		struct list_head *on;
		struct list_head *off;
		int members;
	} watches;

	pthread_mutex_t sentry;
	pthread_cond_t trigger;

	struct list_head coils;
	pthread_t worker; /* coil */

	pthread_t resolver;

	struct list_head clients;
	struct wm_ring *last;
};

struct wm_providers {
	struct list_head others;
	char *module;
	void *handle;
	struct wm_providers_stub *hooks;
	pthread_t worker;
	struct list_head *servers;
	struct wm_local *root;
	int timeout;
	int level;
};

struct wm_servers {
	struct list_head others;
	char *address, *user, *password;
	int port;
	struct hostent hent;
	void *hbuf;
};

struct wm_watches {
	struct list_head others;
	struct list_head *keywords;
	struct wm_links *overlay;
	char *identifier;
	int member;
};

struct wm_links {
	struct wm_watches *watch;
	struct list_head others[];
};

struct wm_keywords {
	struct list_head others;
	char *pattern;
	regex_t compiled;
};

struct wm_grammar_stub {
	struct wm_local *(*parse)(char *, char *);
};

extern void signaler(char *string, void *arg);
extern struct wm_local *parser(char *path, char *name);

extern void init_overlay(struct wm_local *wml, int level);

#endif /* GRAMMAR_H */
