%{ /* -*-C-*- */

/*
 *  $Header: /var/cvsroot/peephole/src/grammar/local/local_lexer.l,v 1.5 2004/03/03 23:08:35 dmr Exp $
 *
 *  Copyright (C) 2003, 2004 Alessandro Di Marco
 */

/*
 *  TODO:
 *
 *  +  reduce the amount of memory allocated with string-tokens; currently a
 *     fixed amount of memory (80 bytes) is allocated for every string-token,
 *     ignoring real token needs.
 */


#include <string.h>

#include "../grammar.h"
#include "local_parser.h"
#include "local.h"

static const char cvsid[]="$Id: local_lexer.l,v 1.5 2004/03/03 23:08:35 dmr Exp $";

#define YY_NO_UNPUT
#define YY_DECL int yylex(YYSTYPE *lvalp, void *param)

%}

%x cmt str inc

%%

%{

static YY_BUFFER_STATE istack[INCLUDE_STACK_SIZE];
static int ibase = 0;

char *newbuf, *oldbuf = NULL;
int newlen, oldlen = 0;

%}

"/*"    {
	BEGIN(cmt);
}

<cmt>{
	[^*]* {
		/* skip */
	}

	"*"+[^*/]* {
		/* skip */
	}

	"*"+"/" {
		BEGIN(INITIAL);
	}
}

"//".* {
	/* skip */
}

"pass" {
	return PASS;
}

"watch" {
	return WATCH;
}

[[:alpha:]][[:alnum:]\-\.\_]* {
	char *identifier;

	identifier = lvalp->string = (char *) malloc(yyleng + 1);
	if (!identifier) {
		yy_fatal_error("no memory for IDENTIFIER");
	}

	strcpy(identifier, yytext);

	return IDENTIFIER;
}

\" {
	oldbuf = (char *) malloc(oldlen = 1);
	if (!oldbuf) {
		yy_fatal_error("no memory for STRING");
	}

	BEGIN(str);
}

<str>{
	\" {
		BEGIN(INITIAL);

		oldbuf[oldlen - 1] = '\0';

		lvalp->string = oldbuf;

		return STRING;
	}

	(\\\\)+|(\\\")+ {
		newbuf = (char *) realloc(oldbuf, newlen = oldlen + (yyleng >> 1));
		if (!newbuf) {
			free(oldbuf);
			yy_fatal_error("no memory for STRING");
		}

		oldbuf = newbuf;

		(void) memset(newbuf + oldlen - 1, yytext[1], yyleng >> 1);

		oldlen = newlen;
	}

	[\ \!\#-\[\]-\~]+ {
		newbuf = (char *) realloc(oldbuf, newlen = oldlen + yyleng);
		if (!newbuf) {
			free(oldbuf);
			yy_fatal_error("no memory for STRING");
		}

		oldbuf = newbuf;

		(void) memcpy(newbuf + oldlen - 1, yytext, yyleng);

		oldlen = newlen;
	}

	\\[^"\\] {
		yy_fatal_error("unknown escape sequence in STRING");
	}

	\n {
		yy_fatal_error("unterminated sequence in STRING");
	}

	. {
		yy_fatal_error("illegal character in STRING");
	}
}

"0" {
	lvalp->number = 0;
	return NATURAL;
}

[1-9][0-9]* {
	lvalp->number = atoi(yytext);
	return NATURAL;
}

[ \t\n]+ {
	/* skip */
}

. {
	return yytext[0];
}

"#include"[ \t]+\< {
	if (ibase >= INCLUDE_STACK_SIZE) {
		yy_fatal_error("includes nested too deeply");
        }

	oldbuf = (char *) malloc(oldlen = 1);
	if (!oldbuf) {
		yy_fatal_error("no memory for INCLUDE");
	}

	BEGIN(inc);
}

<inc>{
	\> {
	        struct wm_config *wmc = param;
                int len;

		len = 1 + strlen(wmc->path);

		newbuf = (char *) realloc(oldbuf, newlen = oldlen + len);
		if (!newbuf) {
			free(oldbuf);
			yy_fatal_error("no memory for INCLUDE");
		}

                memmove(newbuf + len, newbuf, oldlen);

		oldbuf = newbuf;

		(void) memcpy(newbuf, wmc->path, len - 1);

		oldlen = newlen;

		newbuf[len - 1] = '/';

		BEGIN(INITIAL);

		oldbuf[oldlen - 1] = '\0';

		istack[ibase++] = YY_CURRENT_BUFFER;

		yyin = fopen(oldbuf, "r");
		if (!yyin) {
			fprintf(stderr, "file not found: %s", oldbuf);
			yy_fatal_error("\n");
		}

                yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));

		free(oldbuf);
	}

	(\\\\)+|(\\\<)+|(\\\>)+ {
		newbuf = (char *) realloc(oldbuf, newlen = oldlen + (yyleng >> 1));
		if (!newbuf) {
			free(oldbuf);
			yy_fatal_error("no memory for INCLUDE");
		}

		oldbuf = newbuf;

		(void) memset(newbuf + oldlen - 1, yytext[1], yyleng >> 1);

		oldlen = newlen;
	}

	[\ -\;\=\?-\[\]-\~]+ {
		newbuf = (char *) realloc(oldbuf, newlen = oldlen + yyleng);
		if (!newbuf) {
			free(oldbuf);
			yy_fatal_error("no memory for INCLUDE");
		}

		oldbuf = newbuf;

		(void) memcpy(newbuf + oldlen - 1, yytext, yyleng);

		oldlen = newlen;
	}

	\\[^\<\>\\] {
		yy_fatal_error("unknown escape sequence in INCLUDE");
	}

	\n {
		yy_fatal_error("unterminated sequence in INCLUDE");
	}

	. {
		yy_fatal_error("illegal character in INCLUDE");
	}
}

<<EOF>> {
	if (--ibase < 0) {
		yyterminate();
	} else {
                yy_delete_buffer(YY_CURRENT_BUFFER);
                yy_switch_to_buffer(istack[ibase]);
	}
}

%%

int yywrap(void)
{
	return 1;
}
