%{

/*
 *  $Header: /var/cvsroot/peephole/src/grammar/local/local_parser.y,v 1.8 2004/04/19 10:20:30 dmr Exp $
 *
 *  Copyright (C) 2003, 2004 Alessandro Di Marco
 */

/*
 *  TODO:
 */

static const char cvsid[]="$Id: local_parser.y,v 1.8 2004/04/19 10:20:30 dmr Exp $";

%}

%pure_parser

%union {
	struct wm_local *config;
	struct list_head *list;
	char *string;
	int number;
}

%token <string> IDENTIFIER STRING
%token <number> NATURAL
%token WATCH PASS

%type <list> providers servers watches keywords
%type <config> local
%type <string> pass

%{

#include <stdio.h>
#include <stdlib.h>

#include "../grammar.h"
#include "local.h"

int yylex(YYSTYPE *lvalp, void *param);
int yyerror(const char *desc);

#define YYPARSE_PARAM param
#define YYLEX_PARAM param

%}

%%

local:
          /* empty */
		{
			struct wm_config *wmc = param;

			wmc->wml = (struct wm_local *) malloc(sizeof (struct wm_local));
			if (!wmc->wml) {
				return 9;
			}

			wmc->wml->shutdown = 0;
			
			INIT_LIST_HEAD(&wmc->wml->providers);
			INIT_LIST_HEAD(&wmc->wml->watches.main);

			wmc->wml->grade = 0;
			wmc->wml->watches.members = 0;

			wmc->wml->pass = NULL;

			$$ = wmc->wml;
		}
	| local pass
		{
			if ($1->pass) {
				yyerror("already defined password");
				YYERROR;
			}

			$1->pass = $2;

			$$ = $1;
		}
	| local providers
		{
			struct wm_providers *wmp = list_entry($2, struct wm_providers, others);

			wmp->level = $1->grade++;
			wmp->root = $1;

			list_add($2, &$1->providers);

			$$ = $1;
		}
	| local watches
		{
			struct wm_watches *wmw = list_entry($2, struct wm_watches, others);

			wmw->member = $1->watches.members++;

			list_add_tail($2, &$1->watches.main);

			$$ = $1;
		}
	;

pass:
	  PASS '=' STRING ';'
	        {
			$$ = $3;
		}
        ;

providers:
	  IDENTIFIER '(' NATURAL ')' '=' '{' servers '}' ';'
	        {
			struct wm_providers *wmp;

			wmp = (struct wm_providers *) malloc(sizeof (struct wm_providers));
			if (!wmp) {
				return 9;
			}

			wmp->module = $1;

			wmp->timeout = $3;
			wmp->servers = $7;

			$$ = &wmp->others;
                }
	;

servers:
	  '{' STRING ',' NATURAL ',' STRING ',' STRING '}'
	        {
			struct list_head *head;
			struct wm_servers *wms;

			head = (struct list_head *) malloc(sizeof (struct list_head));
			if (!head) {
				return 9;
			}

			INIT_LIST_HEAD(head);

			wms = (struct wm_servers *) malloc(sizeof (struct wm_servers));
			if (!wms) {
				return 9;
			}


			wms->address = $2;
			wms->port = $4;
			wms->user = $6;
			wms->password = $8;

			wms->hbuf = NULL;

			list_add(&wms->others, head);

			$$ = head;
                }
	| servers ','  '{' STRING ',' NATURAL ',' STRING ',' STRING '}'
	        {
			struct wm_servers *wms;

			wms = (struct wm_servers *) malloc(sizeof (struct wm_servers));
			if (!wms) {
				return 9;
			}


			wms->address = $4;
			wms->port = $6;
			wms->user = $8;
			wms->password = $10;

			wms->hbuf = NULL;

			list_add(&wms->others, $1);

			$$ = $1;
                }
	;

watches:
	  WATCH '(' IDENTIFIER ')' '=' '{' keywords '}' ';'
		{
			struct wm_watches *wmw;

			wmw = (struct wm_watches *) malloc(sizeof (struct wm_watches));
			if (!wmw) {
				return 9;
			}

			wmw->identifier = $3;
			wmw->keywords = $7;

			$$ = &wmw->others;
		}
	;

keywords:
	  STRING
		{
			struct list_head *head;
			struct wm_keywords *wmk;

			head = (struct list_head *) malloc(sizeof (struct list_head));
			if (!head) {
				return 9;
			}

			INIT_LIST_HEAD(head);

			wmk = (struct wm_keywords *) malloc(sizeof (struct wm_keywords));
			if (!wmk) {
				return 9;
			}

			wmk->pattern = $1;

			list_add(&wmk->others, head);

			$$ = head;
		}
	| keywords ',' STRING
		{
			struct wm_keywords *wmk;


			wmk = (struct wm_keywords *) malloc(sizeof (struct wm_keywords));
			if (!wmk) {
				return 9;
			}

			wmk->pattern = $3;

			list_add(&wmk->others, $1);

			$$ = $1;
		}
	;

%%

int yyerror(const char *desc)
{
	fprintf(stderr, "%s\n", desc);
	return 0;
}
