/*************************************************************************/
/* peephole.c - peephole main file 					 */
/*  v. 0.1 								 */
/*  Copyright (C) 2004 Alessandro Di Marco - All rights reserved 	 */
/*  by Alessandro Di Marco <dmr@gmx.it>	                 		 */
/* 									 */
/* This program is free software; you can redistribute it and/or 	 */
/* modify  under the terms of the GNU General Public License as 	 */
/* published by the Free Software Foundation; either version 2, or (at 	 */
/* your option) any later version.					 */
/* 									 */
/* This program is distributed in the hope that it will be useful, but 	 */
/* WITHOUT ANY WARRANTY; without even the implied warranty of 		 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 	 */
/* General Public License for more details.				 */
/* 									 */
/* You should have received a copy of the GNU General Public License 	 */
/* along with GNU Emacs; see the file COPYING.  If not, write to the 	 */
/* Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.	 */
/*************************************************************************/

/*
 *  $Header: /var/cvsroot/peephole/src/peephole.c,v 1.23 2004/04/19 10:53:10 dmr Exp $
 *
 *  Copyright (C) 2003, 2004 Alessandro Di Marco
 */

/*
 *  TODO:
 *
 *  +  convert wml->freeze into array of semaphores, one for each level
 *  -  found an easy way to avoid identity flickering
 */

#include <dlfcn.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <pwd.h>
#include <signal.h>
#include <errno.h>

#include "providers/providers.h"
#include "grammar/grammar.h"
#include "peephole.h"

#include "debug.h"

#define FALSE 0
#define TRUE (!FALSE)		/* special reason for this instead of 1 ? */

                                /* yeah! in this way we accomplish future
				 * computers requirements, just in case they
				 * should work in multi level logic ;-)
				 */

static const char cvsid[]="$Id: peephole.c,v 1.23 2004/04/19 10:53:10 dmr Exp $";

/* wm_recv:
 * 	receive a buffer from a socket
 *
 * ARGS:
 *  sfd: file descriptor for used socket
 *
 * RETURNS:
 *  the read buffer
 */

char *wm_recv(int sfd)
{
	char *str, buf[512];
	int len;

	len = recv(sfd, buf, 512, 0);
	if (len < 1) {
		return NULL;
	}

	str = (char *) malloc(--len);
	if (str) {
		memcpy(str, buf, len);
		str[len - 1] = '\0';
	}

	return str;
}

/* wm_send:
 * 	send a buffer through a socket
 *
 * ARGS:
 *  sfd: file descriptor for used socket
 *  str: the buffer to be sent
 *
 * RETURNS:
 *  number of sent bytes, or -1 on error
 */

int wm_send(int sfd, char *str)
{
	char *to;
	int len, err = 0;

	for (len = strlen(str) + 1, to = str + len; str < to; err = send(sfd, str, to - str, 0)) {
		if (err < 0) {
			return err;
		}

		str += err;
	}

	return len;
}

/* build_coil:
 * 	build a coil from given options
 *
 * ARGS:
 *  wml: the config file parse structure
 *
 * RETURNS:
 *  the newly-created coil
 */

struct wm_coil *build_coil(struct wm_local *wml)
{
	struct list_head *pos;
	int i;

	struct wm_coil *wmc;

	wmc = (struct wm_coil *) malloc(sizeof (struct wm_coil) + wml->watches.members * sizeof (struct wm_ring));
	if (!wmc) {
		PRINTD("no memory for coil\n");
		return NULL;
	}

	list_for_each(pos, &wml->watches.main) {
		struct wm_watches *wmw = list_entry(pos, struct wm_watches, others);

		wmc->rings[wmw->member].identifier = wmw->identifier;
		wmc->rings[wmw->member].status = WM_OFF;
	}

	for (i = 0; i < wml->grade; i++) {
		(void) pthread_mutex_lock(&wml->freeze[i]);

		list_for_each(pos, &wml->watches.on[i]) {
			wmc->rings[list_entry(pos, struct wm_links, others[i])->watch->member].status = WM_ON;
		}

		(void) pthread_mutex_unlock(&wml->freeze[i]);
	}

	return wmc;
}

void *coil_worker(void *arg)
{
	struct wm_local *wml = arg;

	struct list_head *pos;

	signal(SIGPIPE, SIG_IGN);

	PRINTD("coil thread created\n");

	while (!wml->shutdown) {
		struct list_head *pos1, *n;
		int i;

		pthread_mutex_lock(&wml->sentry);

		while (list_empty(&wml->coils)) {
			pthread_cond_wait(&wml->trigger, &wml->sentry);
		}

		PRINTD("coil thread woken up\n");

		i = 0;
		list_for_each_safe(pos1, n, &wml->coils) {
			struct wm_coil *wmc = list_entry(pos1, struct wm_coil, others);

			list_del(pos1);

			if (i < 1) {
				i++;

				if (memcmp(wml->last, wmc->rings, wml->watches.members * sizeof (struct wm_ring))) {
					memcpy(wml->last, wmc->rings, wml->watches.members * sizeof (struct wm_ring));

					char buf1[512]; // FIXME: shell code or die... hay!
					char buf2[512]; // FIXME: shell code or die... hay!

					int *vec = (int *) buf2;
					int i, j;

					strcpy(buf1, "");

					for (i = j = 0; j < wml->watches.members; j++) {
						if (wml->last[j].status) {
							strcat(buf1, wml->last[j].identifier);
							strcat(buf1, " ");

							vec[++i] = j;
						}
					}

					if (buf1[0] == '\0') {
						strcat(buf1, "\n");
					} else {
						buf1[strlen(buf1) - 1] = '\n';
					}

					vec[0] = ++i;

					struct list_head *pos2, *n;

					list_for_each_safe(pos2, n, &wml->clients) {
						struct wm_client *wmn = list_entry(pos2, struct wm_client, others);
						int err;

						/* FIXME: check the send return
						 * value and, eventually,
						 * resend the remaining part of
						 * the buffer.
						 */

						err = wmn->mode == PEEP_COOKED ? wm_send(wmn->fd, buf1) : /* PEEP_RAW */ send(wmn->fd, buf2, i * sizeof (int), 0);
						if (err < 0) {
							PRINTD("client connection lost, expunged!\n");
							list_del(&wmn->others);
							close(wmn->fd);
							free(wmn);
						}
					}

					if (list_empty(&wml->clients)) {
						PRINTD("no more clients for user %s, shutting down\n", wml->user);
						wml->shutdown = 1;
					}
				}
			}

			free(wmc);
		}

		pthread_mutex_unlock(&wml->sentry);
	}

	list_del(&wml->others);

	return NULL;
}

#ifdef DEBUG
static inline void debug_watches_list(struct wm_local *wml, int level)
{
	struct list_head *pos;

	(void) pthread_mutex_lock(&wml->freeze[level]);

	if (list_empty(&wml->watches.on[level])) {
		PRINT("EMPTY / ");
	} else {
		list_for_each(pos, &wml->watches.on[level]) {
			struct wm_watches *wmw = list_entry(pos, struct wm_links, others[level])->watch;

			PRINT("%s ", wmw->identifier);
		}

		PRINT("/ ");
	}

	(void) pthread_mutex_unlock(&wml->freeze[level]);
}
#else
static inline void debug_watches_list(struct wm_local *wml, int level) {}
#endif

void *provider_worker(void *arg)
{
	struct wm_providers *wmp = arg;

	struct wm_local *wml = wmp->root;
	struct wm_providers_stub *hooks = wmp->hooks;

	struct list_head *pos;

	FILE *fp;
	int msgs, last;
	int i;

	int timeout;

	PRINTD("%s thread created\n", wmp->module);

	while (!wml->shutdown) {
		init_overlay(wml, wmp->level);

		list_for_each(pos, wmp->servers) {
			struct wm_servers *wms = list_entry(pos, struct wm_servers, others);

			if (wms->hbuf) {
				/* wms->hent is valid, go ahead... */

				fp = hooks->open(&wms->hent, wms->port, wms->user, wms->password);
				if (fp) {
					if (hooks->stat(fp, &msgs, &last) >= 0) {
						for (i = last + 1; i <= msgs; i++) {
							if (hooks->match(fp, i, signaler, wmp) < 0) {
								break;
							}
						}
					}

					hooks->close(fp);

					PRINTD("provider %s: ", wms->address);

					debug_watches_list(wml, wmp->level);

					PRINT("\n");
				} else {
					PRINTD("server %s: doesn't respond\n", wms->address);
				}
			} else {
				PRINTD("%s address unresolved, skipped!\n", wms->address);
			}
		}

		struct wm_coil *wmc;

		PRINTD("creating coil: ");

		int i;

		for (i = 0; i < wml->grade; i++) {
			debug_watches_list(wml, i);
		}

		PRINT("\n");

		wmc = build_coil(wml);

		if (wmc) {
			(void) pthread_mutex_lock(&wml->sentry);

			list_add(&wmc->others, &wml->coils);

			pthread_cond_signal(&wml->trigger);

			(void) pthread_mutex_unlock(&wml->sentry);
		}

		for (timeout = wmp->timeout; timeout > 0; timeout = sleep(timeout));
	}

	PRINTD("%s thread shutdown\n", wmp->module);

	return NULL;
}

static struct hostent *gethostbyname_server(struct wm_servers *wms)
{
	struct hostent *hent = NULL;

	void *hbuf;
	int herr;

	int size = 1 << 9;

	hbuf = malloc(size);
	if (hbuf) {
		while (gethostbyname_r(wms->address, &wms->hent, hbuf, size, &hent, &herr) == ERANGE) {
			PRINTD("insufficient hbuf size, raising to %d\n", size << 1);
			free(hbuf);
			hbuf = malloc(size <<= 1);
			if (!wms->hbuf) {
				break;
			}
		}

		if (hent) {
			wms->hbuf = hbuf;
		} else {
			if (hbuf) {
				free(hbuf);
			}
		}
	}

	return hent;
}

void *resolve_providers(void *arg)
{
	struct wm_local *wml = arg;

	struct list_head *pos, *pos1;
	int unresolved;

	do {
		unresolved = 0;

		list_for_each(pos, &wml->providers) {
			struct wm_providers *wmp = list_entry(pos, struct wm_providers, others);

			list_for_each(pos1, wmp->servers) {
				struct wm_servers *wms = list_entry(pos1, struct wm_servers, others);

				if (wml->shutdown) {
					PRINTD("shutdown acknowledged, resolve quitting\n");
					return NULL;
				}

				if (!wms->hbuf) {
					PRINTD("resolving host '%s'\n", wms->address);

					unresolved++;

					if (gethostbyname_server(wms)) {
						unresolved--;
						PRINTD("host '%s' is at %s\n", wms->address, inet_ntoa(*((struct in_addr *) wms->hent.h_addr)));
					} else {
						PRINTD("cannot resolve host '%s'\n", wms->address);
					}
				} else {
					PRINTD("host %s already resolved at %s\n", wms->address, inet_ntoa(*((struct in_addr *) wms->hent.h_addr)));
				}
			}
		}

		if (unresolved > 0) {
			PRINTD("cannot resolve %d hosts...\n", unresolved);
			sleep(RESOLVE_TIMEOUT);
		}

	} while(unresolved > 0);

	PRINTD("all hosts have been resolved, quitting\n");

	return NULL;
}

int startup_providers(struct wm_local *wml)
{
	struct list_head *pos, *pos1;
	pid_t pid;

	if (pthread_create(&wml->resolver, NULL, resolve_providers, wml)) {
		PRINTD("cannot create resolver thread\n");
		return -EFAULT;
	}

	sleep(RESOLVE_WARMUP);

	list_for_each(pos, &wml->providers) {
		struct wm_providers *wmp = list_entry(pos, struct wm_providers, others);
		char buf[512]; // FIXME: beautiful shell code entry point!

		strcpy(buf, LIBPATH);
		strcat(buf, "/");
		strcat(buf, wmp->module);
		strcat(buf, ".so");

		wmp->handle = dlopen(buf, RTLD_NOW);
		if (!wmp->handle) {
			PRINTD("%s\n", dlerror());
			return -EFAULT;
		}

		wmp->hooks = (struct wm_providers_stub *) dlsym(wmp->handle, "hooks");
		if (!wmp->hooks) {
			PRINTD("%s\n", dlerror());
			return -EFAULT;
		}

		if (pthread_create(&wmp->worker, NULL, provider_worker, wmp)) {
			PRINTD("cannot create worker threads\n");
			/* FIXME: hard exit... please clean me up! UPDATE: Fix *all*
			 * 'if' body (what about perror or stderr?) */
			return -EFAULT;
		}
	}

	return 0;
}

int main(int argc, char** argv)
{
	LIST_HEAD(world);

	struct sockaddr_in laddr, raddr;
	int sfd, nfd;
	int size;

	int yes = 1;
	int err;

	if (daemon(1, 1) < 0) {
		FATAL("(trying to daemonize)");
	}

	PRINTD("peephole start\n");

	signal(SIGPIPE, SIG_IGN);

	sfd = socket(AF_INET, SOCK_STREAM, 0);
	if (!~sfd) {
		FATAL("(creating socket)");
	}

	err = setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (int));
	if (!~err) {
		FATAL("(setting sock opts)");
	}

	laddr.sin_family = AF_INET;
	laddr.sin_port = htons(MYPORT);
	laddr.sin_addr.s_addr = MYADDRESS;

	memset(&laddr.sin_zero, '\0', 8);

	err = bind(sfd, (struct sockaddr *) &laddr, sizeof (struct sockaddr));
	if (!~err) {
		FATAL("(binding)");
	}

	err = listen(sfd, BACKLOG);
	if (!~err) {
		FATAL("(listening)");
	}

	while (1) {
		struct list_head *pos, *n;
		char *user, *pass;

		size = sizeof (struct sockaddr_in);

		nfd = accept(sfd, (struct sockaddr *) &raddr, &size);
		if (!~nfd) {
			perror("accept");
			continue;
		}

		if (wm_send(nfd, "\n" WM_VERSION "\n\n") < 1) {
			close(nfd);
			continue;
		}

		user = wm_recv(nfd);
		pass = wm_recv(nfd);

		if (!user || !pass || strncmp(user, "user", 4) || strncmp(pass, "pass", 4)) {
			close(nfd);
			continue;
		}

		struct passwd *pw = getpwnam(&user[5]);

		if (!pw) {
			PRINTD("user not found: %s\n", &user[5]);
			close(nfd);
			continue;
		}

		struct wm_local *wml = NULL;

		list_for_each(pos, &world) {
			wml = list_entry(pos, struct wm_local, others);

			if (!strcmp(wml->user, &user[5])) {
				break;
			}

			wml = NULL;
		}

		if (wml) {
			if (!strcmp(wml->pass, &pass[5])) {
				struct list_head *pos;
				char buf[512]; // FIXME: shell code...
				int j;

				strcpy(buf, "");

				list_for_each(pos, &wml->watches.main) {
					struct wm_watches *wmw = list_entry(pos, struct wm_watches, others);

					strcat(buf, wmw->identifier);
					strcat(buf, " ");
				}

				buf[strlen(buf) - 1] = '\n';

				if (wm_send(nfd, buf) < 1) {
					close(nfd);
					continue;
				}

				struct wm_client *wmn;

				wmn = (struct wm_client *) malloc(sizeof (struct wm_client));
				if (!wmn) {
					perror("no memory for new client\n");
					continue;
				}

				wmn->fd = nfd;

				char *mode = wm_recv(nfd);

				if (mode && !strncmp(mode, "raw", 3)) {
					wmn->mode = PEEP_RAW;
				} else if (mode && !strncmp(mode, "cooked", 6)) {
					wmn->mode = PEEP_COOKED;
				} else {
					close(nfd);
					free(wmn);
					continue;
				}

				if (wmn->mode == PEEP_COOKED) {
					strcpy(buf, "");

					pthread_mutex_lock(&wml->sentry);

					for (j = 0; j < wml->watches.members; j++) {
						if (wml->last[j].status) {
							strcat(buf, wml->last[j].identifier);
							strcat(buf, " ");
						}
					}

					pthread_mutex_unlock(&wml->sentry);

					if (buf[0] == '\0') {
						strcat(buf, "\n");
					} else {
						buf[strlen(buf) - 1] = '\n';
					}

					if (wm_send(nfd, buf) < 1) {
						close(nfd);
						free(wmn);
						continue;
					}
				} else { /* PEEP_RAW */
					int *vec = (int *) buf;
					int i;

					pthread_mutex_lock(&wml->sentry);

					for (i = j = 0; j < wml->watches.members; j++) {
						if (wml->last[j].status) {
							vec[++i] = j;
						}
					}

					pthread_mutex_unlock(&wml->sentry);

					vec[0] = ++i;

					/* FIXME: check the return value and,
					 * eventually, resend the remaining
					 * part of the buffer.
					 */

					(void) send(nfd, buf, i * sizeof (int), 0);
				}

				pthread_mutex_lock(&wml->sentry);

				list_add(&wmn->others, &wml->clients);

				pthread_mutex_unlock(&wml->sentry);
			} else {
				PRINTD("wrong password\n");
				close(nfd);
			}

			continue;
		}

		wml = parser(pw->pw_dir, ".peepholerc");
		if (!wml) {
			PRINTD("%s: config not found!\n", pw->pw_dir);
			close(nfd);
			continue;
		}

		wml->user = &user[5];

		if (!strcmp(wml->pass, &pass[5])) {
			struct list_head *pos;
			char buf[512]; // FIXME: shell code...

			strcpy(buf, "");

			list_for_each(pos, &wml->watches.main) {
				struct wm_watches *wmw = list_entry(pos, struct wm_watches, others);

				strcat(buf, wmw->identifier);
				strcat(buf, " ");
			}

			buf[strlen(buf) - 1] = '\n';

			if (wm_send(nfd, buf) < 1) {
				close(nfd);
				continue;
			}

			struct wm_client *wmn;

			wmn = (struct wm_client *) malloc(sizeof (struct wm_client));
			if (!wmn) {
				perror("no memory for new client");
				close(nfd);
				continue;
			}

			wmn->fd = nfd;

			char *mode = wm_recv(nfd);

			if (mode && !strncmp(mode, "raw", 3)) {
				wmn->mode = PEEP_RAW;
			} else if (mode && !strncmp(mode, "cooked", 6)) {
				wmn->mode = PEEP_COOKED;
			} else {
				close(nfd);
				free(wmn);
				continue;
			}

			pthread_mutex_lock(&wml->sentry);

			list_add(&wmn->others, &wml->clients);

			pthread_mutex_unlock(&wml->sentry);
		} else {
			/* FIXME: dispose wm_local... */
			PRINTD("wrong password\n");
			close(nfd);
			continue;
		}

		err = startup_providers(wml);
		if (err < 0) {
			close(nfd);
			continue;
		}

		if (pthread_create(&wml->worker, NULL, coil_worker, wml)) {
			PRINTD("cannot create feeder thread\n");
			return -EFAULT;
		}

		list_add(&wml->others, &world);
	}

	close(sfd);

	return 0;
}
