/*******************************************************************************/
/* apop.c - apop module for peephole 					       */
/*  v. 0.0.0 								       */
/*  Copyright (C) 2004 Paolo Ardoino, Alessandro Di Marco		       */
/*  by Paolo Ardoino  <ardoino.gnu@disi.unige.it> 			       */
/*  									       */
/* This program is free software; you can redistribute it and/or 	       */
/* modify it under the terms of the GNU General Public License as 	       */
/* published by the Free Software Foundation; either version 2 of the 	       */
/* License, or (at your option) any later version.			       */
/* 									       */
/* This program is distributed in the hope that it will be useful, but 	       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of 		       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 	       */
/* General Public License for more details.				       */
/*  									       */
/* You should have received a copy of the GNU General Public License 	       */
/* along with this program; if not, write to the Free Software 		       */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-	       */
/* 1307, USA.								       */
/*******************************************************************************/

/*
 *  $Header: /var/cvsroot/peephole/src/providers/apop/apop.c,v 1.9 2004/04/19 10:24:42 dmr Exp $
 *
 */

/*
 *  TODO:
 */

#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <openssl/md5.h>
#include <openssl/evp.h>

#include "../providers.h"

#include "apop.h"

#include "../../debug.h"

static const char cvsid[]="$Id: apop.c,v 1.9 2004/04/19 10:24:42 dmr Exp $";

static void bin2hex(unsigned char *inbuf, unsigned char *outbuf)
{
	const char *hex = "0123456789abcdef";
	int ctr;

	for (ctr = 0; ctr < 16; ctr++, inbuf++) {
		*outbuf++ = hex[*inbuf >> 4];
		*outbuf++ = hex[*inbuf & 0xf];
	}

	*outbuf = '\0';
}

static void md5_hash(char *inbuf)
{
	EVP_MD_CTX ctx;
	unsigned char tmpbuf[BUFFER_SIZE];

	strcpy(tmpbuf, (unsigned char *) inbuf);

	EVP_DigestInit(&ctx,EVP_md5());
	EVP_DigestUpdate(&ctx, tmpbuf, strlen(tmpbuf));
	EVP_DigestFinal(&ctx, tmpbuf, NULL);

	bin2hex(tmpbuf, inbuf);
}

static void *open_apop(struct hostent *hent, int port, char *user, char *password)
{
	struct sockaddr_in sa;

	int sd;
	FILE *sf;

	char cmd[BUFFER_SIZE], md5h[BUFFER_SIZE];
	char *ps=NULL;
	char *pe=NULL;

	sa.sin_family = hent->h_addrtype;
	sa.sin_port = htons(port);

	memcpy((char *) &sa.sin_addr.s_addr, hent->h_addr_list[0], hent->h_length);

	if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		PRINTD("cannot open socket for '%s'\n", hent->h_name);
		return NULL;
	}

	PRINTD("connecting to '%s'\n", hent->h_name);

	if (connect(sd, (struct sockaddr *) &sa, sizeof(struct sockaddr_in)) == -1) {
		PRINTD("connection failed on '%s'\n", hent->h_name);
		close(sd);
		return NULL;
	}

	sf = fdopen(sd, "r+");

	PRINTD("'%s' connected\n", hent->h_name);

	READ_BUFFER(cmd);
	if (OK_BUFFER(cmd)) {
		strcpy(md5h, cmd);

		if ((ps = strchr(md5h, '<')) != NULL && (pe = strrchr(md5h, '>')) != NULL && pe > ps) {
			*(pe + 1) = '\0';

			strcat(ps, password);
			md5_hash(ps);

			sprintf(cmd, "APOP %s %s\r\n", user, ps);
			WRITE_BUFFER(cmd);

			READ_BUFFER(cmd);
			if (OK_BUFFER(cmd)){
				READ_BUFFER(cmd);
				return sf;
			}
		} else {
			PRINTD("apop not detected on '%s'\n", hent->h_name);
			fclose(sf);

			return NULL;
		}
	}

	PRINTD("cannot autenthicate on '%s'\n", hent->h_name);
	fclose(sf);

	return NULL;
}

static void close_apop(void *sf)
{
	char cmd[BUFFER_SIZE];

	WRITE_BUFFER("QUIT");
	READ_BUFFER(cmd);
	if (ERR_BUFFER(cmd)) {
		PRINTD("apop refuses to quit\n");
	}

	fclose(sf);
}

static int stat_apop(void *sf, int *msgs, int *last)
{
	char cmd[BUFFER_SIZE];

	WRITE_BUFFER("STAT");
	READ_BUFFER(cmd);
	if (ERR_BUFFER(cmd)) {
		PRINTD("error receiving stat\n");
		return -EFAULT;
	}

	*msgs = atoi(&cmd[4]);

	WRITE_BUFFER("LAST");
	READ_BUFFER(cmd);

	*last = OK_BUFFER(cmd) ? atoi(&cmd[4]) : 0;

	return 0;
}

static int match_apop(void *sf, int msg, signaler_cb signaler, void *arg)
{
	char cmd[50 * BUFFER_SIZE];
	int pos;

	sprintf(cmd, "TOP %d 0", msg);
	WRITE_BUFFER(cmd);
	READ_BUFFER(cmd);
	if (ERR_BUFFER(cmd)) {
		PRINTD("message %d not found\n", msg);
		return -EFAULT;
	}

	/*  FIXME: cmd boundary is unchecked here! */
	for (pos = 0; fgets(cmd + pos, BUFFER_SIZE, sf); pos += strlen(cmd + pos)) {
		if (!strncmp(cmd + pos, ".", 1)) {
			signaler(cmd, arg);

			return 0;
		}
	}

	PRINTD("troubles in message %d reception\n", msg);

	return -EFAULT;
}

struct wm_providers_stub hooks = {
	open_apop,
	close_apop,
	stat_apop,
	match_apop
};
