/*
 *  $Header: /var/cvsroot/peephole/src/providers/pop3/pop3.c,v 1.8 2004/04/19 10:24:42 dmr Exp $
 *
 *  Copyright (C) 2003, 2004 Alessandro Di Marco
 */

/*
 *  TODO:
 */

#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "../providers.h"

#include "pop3.h"

#include "../../debug.h"

static const char cvsid[]="$Id: pop3.c,v 1.8 2004/04/19 10:24:42 dmr Exp $";

static void *open_pop3(struct hostent *hent, int port, char *user, char *password)
{
	struct sockaddr_in sa;
	int sd;

	FILE *sf;
	char cmd[BUFFER_SIZE];

	int err;

	sa.sin_family = hent->h_addrtype;
	sa.sin_port = htons(port);

	memcpy((char *) &sa.sin_addr.s_addr, hent->h_addr_list[0],
	       hent->h_length);

	sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sd < 0) {
		PRINTD("cannot open socket for '%s'\n", hent->h_name);
		return NULL;
	}

	PRINTD("connecting to '%s'\n", hent->h_name);

	err = connect(sd, (struct sockaddr *) &sa, sizeof (struct sockaddr_in));
	if (err < 0) {
		PRINTD("connection failed on '%s'\n", hent->h_name);
		close(sd);
		return NULL;
	}

	sf = fdopen(sd, "r+");

	PRINTD("'%s' connected\n", hent->h_name);

	READ_BUFFER(cmd);
	if (OK_BUFFER(cmd)) {
		sprintf(cmd, "USER %s", user);
		WRITE_BUFFER(cmd);
		READ_BUFFER(cmd);
		if (OK_BUFFER(cmd)) {
			sprintf(cmd, "PASS %s", password);
			WRITE_BUFFER(cmd);
			READ_BUFFER(cmd);
			if (OK_BUFFER(cmd)) {
				return sf;
			}
		}
	}

	PRINTD("cannot authenticate on '%s'\n", hent->h_name);
	fclose(sf);

	return NULL;
}

static void close_pop3(void *sf)
{
	char cmd[BUFFER_SIZE];

	WRITE_BUFFER("QUIT");
	READ_BUFFER(cmd);
	if (ERR_BUFFER(cmd)) {
		PRINTD("pop3 refuses to quit\n");
	}

	fclose(sf);
}

static int stat_pop3(void *sf, int *msgs, int *last)
{
	char cmd[BUFFER_SIZE];

	WRITE_BUFFER("STAT");
	READ_BUFFER(cmd);
	if (ERR_BUFFER(cmd)) {
		PRINTD("error receiving stat\n");
		return -EFAULT;
	}

	*msgs = atoi(&cmd[4]);

	WRITE_BUFFER("LAST");
	READ_BUFFER(cmd);

	*last = OK_BUFFER(cmd) ? atoi(&cmd[4]) : 0;

	return 0;
}

static int match_pop3(void *sf, int msg, signaler_cb signaler, void *arg)
{
	char cmd[50 * BUFFER_SIZE];
	int pos;

	sprintf(cmd, "TOP %d 0", msg);

	WRITE_BUFFER(cmd);
	READ_BUFFER(cmd);
	if (ERR_BUFFER(cmd)) {
		PRINTD("message %d not found\n", msg);
		return -EFAULT;
	}

	/*  FIXME: cmd boundary is unchecked here! */
	for (pos = 0; fgets(cmd + pos, BUFFER_SIZE, sf); pos += strlen(cmd + pos)) {
		if (!strncmp(cmd + pos, ".", 1)) {
			signaler(cmd, arg);
			return 0;
		}
	}

	PRINTD("troubles in message %d reception\n", msg);

	return -EFAULT;
}

struct wm_providers_stub hooks = {
	open_pop3,
	close_pop3,
	stat_pop3,
	match_pop3
};
